package solutions.bitbadger.documents.query

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.*
import kotlin.test.assertEquals

/**
 * Unit tests for the `RemoveFields` object
 */
@DisplayName("RemoveFields (Query)")
class RemoveFieldsTest {

    /** Test table name */
    private val tbl = "test_table"

    /**
     * Reset the dialect
     */
    @AfterEach
    fun cleanUp() {
        Configuration.dialectValue = null
    }

    @Test
    @DisplayName("byId generates correctly (PostgreSQL)")
    fun byIdPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "UPDATE $tbl SET data = data - :name::text[] WHERE data->>'id' = :id",
            RemoveFields.byId<String>(tbl, Parameters.fieldNames(listOf("a", "z"))),
            "Remove Fields query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byId generates correctly (SQLite)")
    fun byIdSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals(
            "UPDATE $tbl SET data = json_remove(data, :name0, :name1) WHERE data->>'id' = :id",
            RemoveFields.byId<String>(tbl, Parameters.fieldNames(listOf("a", "z"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byFields generates correctly (PostgreSQL)")
    fun byFieldsPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "UPDATE $tbl SET data = data - :name::text[] WHERE data->>'f' > :g",
            RemoveFields.byFields(tbl, Parameters.fieldNames(listOf("b", "c")), listOf(Field.greater("f", "", ":g"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byFields generates correctly (SQLite)")
    fun byFieldsSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals(
            "UPDATE $tbl SET data = json_remove(data, :name0, :name1) WHERE data->>'f' > :g",
            RemoveFields.byFields(tbl, Parameters.fieldNames(listOf("b", "c")), listOf(Field.greater("f", "", ":g"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byContains generates correctly (PostgreSQL)")
    fun byContainsPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "UPDATE $tbl SET data = data - :name::text[] WHERE data @> :criteria",
            RemoveFields.byContains(tbl, Parameters.fieldNames(listOf("m", "n"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byContains fails (SQLite)")
    fun byContainsSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertThrows<DocumentException> { RemoveFields.byContains(tbl, listOf()) }
    }

    @Test
    @DisplayName("byJsonPath generates correctly (PostgreSQL)")
    fun byJsonPathPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "UPDATE $tbl SET data = data - :name::text[] WHERE jsonb_path_exists(data, :path::jsonpath)",
            RemoveFields.byJsonPath(tbl, Parameters.fieldNames(listOf("o", "p"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byJsonPath fails (SQLite)")
    fun byJsonPathSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertThrows<DocumentException> { RemoveFields.byJsonPath(tbl, listOf()) }
    }
}